"""
Основная фаза симплекс-метода.
"""

import copy
import numpy as np

from rsksymplexmethod import utils


# the max number of iterations to attempt:
MAX_ITER = 42


def iteration(c, A, x, B, Ab_inv_prev, index):
    """
    Итерация основной фазы симплекс-метода.

    INPUT:
    - c
    - A
    - x
    - B
    - Ab_inv_prev: матрица обратная матрице Ab из предыдущей итерации
      (на первой итерации - None)
    - index: the index of the element of the B vector that got changed
      in the previous iteration

    OUTPUT: (iteration bundle tuple)
    - x
    - B
    - unbound
    - solved
    - Ab_inv: матрица обратная Ab с текущей итерации
    - index: the index of the element of the B vector that got changed
      in this iteration
    """
    n = len(c)

    nB = utils.list_diff(B, list(range(n)))
    Ab = A[:, B]

    if Ab_inv_prev is None and index is None:
        Ab_inv = np.linalg.inv(Ab)
    else:
        Ab_inv = utils.smart_invertion(Ab_inv_prev, Ab[:, index], index)

    cb = [c[i] for i in B]
    u = cb @ Ab_inv
    delta = u @ A - c
    nB_delta = [el for i, el in enumerate(delta) if i in nB]

    if all([el >= 0 for el in nB_delta]):
        return x, B, False, True, None, None

    j0 = 0
    for i, v in enumerate(delta):
        if v < 0:
            j0 = i
            break

    z = Ab_inv @ A[:, j0]
    theta = np.array([x[j]/z_i if z_i > 0 else np.inf
                      for z_i, j in zip(z, B)])

    theta0 = np.amin(theta)
    if theta0 == np.inf:
        return x, B, True, True, None, None
    theta0_index = np.where(theta == theta0)[0][0] # np.where returns a tuple hence the indexing

    j_ast = B[theta0_index]
    j_ast_index = np.where(np.array(B) == j_ast)[0][0]

    B[j_ast_index] = j0
    nB = utils.list_diff(B, list(range(n)))
    for i in nB:
        x[i] = 0
    x[j0] = theta0
    for j_index, j in enumerate(B):
        if j != j0:
            x[j] -= theta0*z[j_index]
    
    return x, B, False, False, Ab_inv, j_ast_index


def run(c, A, x, B):
    """
    The main phase algorithm of the custom 2-phase symplex method.
    
    INPUT:
    - c: np.array: вектор стоимостей из целевой функции
    - A: 2d np.array: матрица ограничений
    - x: np.array: начальный базисный допустимый план
    - B: множество базисных индексов

    OUTPUT: (main phase bundle dict)
    - iter_num: number of iterations made
    - unbound: True, если целевая функция не ограничена сверху на
      множестве допустимых планов
    - solved: True, если решение ЗЛП найдено
    - x: итоговый допустимый план
    - B: итоговое множество базисных индексов
    """

    # the main phase is run only once or twice for a LPP
    # hence the copying wouldn't affect the speed much
    c = copy.deepcopy(c)
    A = copy.deepcopy(A)
    x = copy.deepcopy(x)
    B = copy.deepcopy(B)

    n = len(c)
    m = len(A)

    # some assertions could be made here for the reliability's sake

    Ab_inv = None
    index = None
    for i in range(MAX_ITER):
        x, B, unbound, solved, Ab_inv, index = iteration(c, A, x, B, Ab_inv, index)
        if unbound:
            return {
                'iter_num': i+1,
                'unbound': True,
                'solved': True,
                'x': x,
                'B': B,
            }
        if solved:
            return {
                'iter_num': i+1,
                'unbound': False,
                'solved': True,
                'x': x,
                'B': B,
            }

    # the "I give up" result
    return {
        'iter_num': i+1,
        'unbound': False,
        'solved': False,
        'x': x,
        'B': B,
    }



"""
This is the module whose functionality is to be used from the outside
of the rsksymplexmethod package. It provides the functions to solve LPPs
both using the scipy method, and the custom method implementing the
2-phase algorithm of solving LPPs.
"""

import numpy as np
from scipy.optimize import linprog

from rsksymplexmethod import initial
from rsksymplexmethod import main


def custom_solve(c, A, b):
    """
    Полный алгоритм для симплекс метода.

    INPUT: ЗЛП в канонической форме:
    - c: вектор стоимостей
    - A: матрица ограничений
    - b

    OUTPUT: symplex method bundle dict:
    - solved: True if the solution is found
    - inconsistent: True if LPP is found to be inconsistent
    - unbound: True если обнаружилось, что ц.ф. ЗЛП не ограничена
      сверху на множестве допустимых планов
    - x: либо оптимальный план ЗЛП, либо None
    - details (optional): примечания
    """
    c = np.array(c)
    A = np.array(A)
    b = np.array(b)
    
    init_phase_bundle = initial.run(c, A, b)

    if not init_phase_bundle['success']:
        return {
            'solved': False,
            'details': 'initial phase failure; '
            'inner main phase algorithm reached iteration number limit'
        }

    # process the initial phase result:
    x = init_phase_bundle['x']
    B = init_phase_bundle['B']
    
    print(x)
    print(B)

    main_phase_bundle = main.run(c, A, x, B)

    if not main_phase_bundle['solved']:
        return {
            'solved': False,
            'details': 'main phase failure; '
            'main phase algorithm reached iteration number limit'
        }

    # process the main phase result:
    return {
        'solved': True,
        'x': main_phase_bundle['x'],
    }


def scipy_solve(c, A, b):
    """
    Solve an LPP using scipy.
    Return what scipy returns and let the user handle that themself.
    """
    return linprog(-c, A_eq=A, b_eq=b)



"""
Some additional functions for the symplex method.
"""

import copy
import numpy as np


def list_diff(a, b):
    """
    Get 2 lists difference (as if they are sets)
    """
    return list(set(a) - set(b)) + list(set(b) - set(a))


def replace_col(A, b, index):
    """
    Replace A matrix column #index by the b vector.
    """
    assert len(A) == len(b)
    n = len(A)
    m = len(A[0])
    for i in range(n):
        assert len(A[i]) == m
        A[i][index] = b[i]


def smart_invertion(A_inv, x, i):
    """
    Get A matrix inverted, x vector and i index.
    Return A dash matrix inverted, or None if not invertable.
    """
    n = len(A_inv)

    # 1:
    l = A_inv @ x
    if l[i] == 0:
        return None

    # 2:
    l_wave = copy.deepcopy(l)
    l_wave[i] = -1

    # 3:
    l_hat = -(1/l[i]) * l_wave

    # 4:
    Q = np.identity(n)
    replace_col(Q, l_hat, i)

    # 5:
    A_dash_inv = optimul(Q, A_inv, i)
    return A_dash_inv


def optimul(A, B, index):
    # TODO: some checks and assertions should be here

    n = len(A)

    C = np.zeros([n, n])
    for i in range(n):
        for j in range(n):
            C[i][j] += A[i][index] * B[index][j]
            if i != index:
                C[i][j] += B[i][j]

    return C



"""
The initial phase of the custom 2-phase symplex method of solving LPPs.
"""

import numpy as np
from rsksymplexmethod import main
from rsksymplexmethod import utils


def correcting_algorithm(c_wave, A, b, A_wave, B):
    """
    Корректирующий алгоритм.

    # TODO: возможно стоит добавить лимит на количество итераций

    INPUT:
    - c_wave
    - A
    - b
    - A_wave
    - B

    OUTPUT: (correcting algorithm bundle tuple)
    - A
    - b
    - B
    """
    m = len(A_wave)
    n = len(A_wave[0]) - m

    while True:
        # если в B есть такой индекс jk, что jk > n, ...
        found_jk = False
        for k, jk in enumerate(B):
            if jk >= n:
                found_jk = True
                break
        
        if not found_jk:
            break

        # ... то для каждого небазисного индекса j < n вычислить
        # вектор l[j] = A_wave_b_inv @ A_wave_j
        i = jk - n
        nB = [j for j in range(n) if j not in B]
        A_wave_b_inv = np.linalg.inv(A_wave[:, B])
        l = [None]*n
        for j in nB:
            A_wave_j = A_wave[:, j]
            l[j] = A_wave_b_inv @ A_wave_j

        # 2 случая:
        # 1: пусть найдется индекс j (небазисный) родной переменной,
        #    такой, что k-ая компонента его вектора l не равна нулю:
        found_j = False
        for j, lj in enumerate(l):
            if lj is not None:
                if lj[k] != 0:
                    found_j = True
                    B[k] = j
                    break

        # 2: пусть такого индекса нет:
        if not found_j:
            # 1: из B удалить jk:
            B = np.delete(B, k)
            # 2: из A&b удалить ограничение #i:
            A = np.delete(A, i, 0)
            b = np.delete(b, i)
            # 3: из A_wave удалить ограничение #i:
            A_wave = np.delete(A_wave, i, 0)
            # 4: из A_wave&c_wave удалить столбец/компоненту #(n+i):
            A_wave = np.delete(A_wave, n+i, 1)
            c_wave = np.delete(c_wave, n+i)

    return A, b, B


def run(c, A, b):
    """
    The algorithm for the initial phase.

    INPUT:
    - c
    - A
    - b

    OUTPUT: (init phase bundle dict)
    - success: True if this stage was successful (rn it indicates
      whether the inner main phase algorithm call was successful
      and yielded the solution to вспомогательная ЗЛП)
    - infeasible: True if the LPP is found to be infeasible
    - A: new A
    - b: new b
    - x: начальный базисный допустимый план для входной ЗЛП
    - B: множество базисных индексов
    """
    m = len(A)
    n = len(A[0])

    # 1:
    for i in range(m):
        if b[i] < 0:
            b[i] *= -1
            A[i] = [-a for a in A[i]]

    # 2:
    c_wave = np.array([0]*n + [-1]*m)
    Em = np.identity(m)
    A_wave = np.hstack((A, Em))
    x_hat = np.concatenate((np.zeros(n), b))
    B_hat = [n+i for i in range(m)]

    # 3:
    extra_task_solution_bundle = main.run(c_wave, A_wave, x_hat, B_hat)
    etsb = extra_task_solution_bundle

    # some error handling:
    if not etsb['solved']:
        return {
            'success': False,
        }

    y = etsb['x']
    B = etsb['B']

    if any(y[n:]) != 0:
        return {
            'success': True,
            'infeasible': True,
        }

    x_ast = y[:n]
    A, b, B = correcting_algorithm(c_wave, A, b, A_wave, B)

    return {
        'success': True,
        'infeasible': False,
        'A': A,
        'b': b,
        'x': x_ast,
        'B': B,
    }



"""
Двойственный симплекс-метод.
"""

import copy
import numpy as np

from rsksymplexmethod import utils


# the max number of iterations to attempt:
MAX_ITER = 42


def iteration(c, A, b, B, Ab_inv_prev, index):
    """
    Итерация двойственного симплекс-метода.

    INPUT:
    - c
    - A
    - b
    - B
    - Ab_inv_prev: матрица обратная матрице Ab из предыдущей итерации
      (на первой итерации - None)
    - index: the index of the element of the B vector that got changed
      in the previous iteration

    OUTPUT: (iteration bundle tuple)
    - solved: bool
    - infeasible: bool
    - kappa
    - y
    - Ab_inv: Ab_inv calculated in this iteration
    - index: index calculated in this iteration

    All changes to the input variables are made by reference.
    """
    n = len(c)

    nB = utils.list_diff(B, list(range(n)))
    Ab = A[:, B]

    if Ab_inv_prev is None and index is None:
        Ab_inv = np.linalg.inv(Ab)
    else:
        Ab_inv = utils.smart_invertion(Ab_inv_prev, Ab[:, index], index)

    cb = [c[i] for i in B]
    y = cb @ Ab_inv
    kappa = np.zeros(n)
    kappa_B = Ab_inv @ b
    for Bi, kappa_Bi in zip(B, kappa_B):
        kappa[Bi] = kappa_Bi

    if all([el >= 0 for el in kappa]):
        return True, False, kappa, y, None, None

    for i, el in enumerate(kappa):
        if el < 0:
            kappa_ast_index = i
            break

    for i, el in enumerate(B):
        if el == kappa_ast_index:
            kappa_ast_index_B_index = i # oh boy
            break

    delta_y = Ab_inv[kappa_ast_index_B_index]
    
    mu = {}
    for j in nB:
        mu[j] = delta_y @ A[:, j]

    if all([el >= 0 for el in list(mu.values())]):
        return True, True, kappa, y, None, None

    sigma = {}
    for i in nB:
        if mu[i] < 0:
            sigma[i] = (c[i] - A[:,i]@y) / mu[i]

    sigma0_index = list(sigma.keys())[0]
    for i, s in sigma.items():
        if s < sigma[sigma0_index]:
            sigma0_index = i
            break

    for i, Bi in enumerate(B):
        if Bi == kappa_ast_index:
            B[i] = sigma0_index
            new_index = i
            break
    
    return False, False, None, None, Ab_inv, new_index


def run(c, A, b, B):
    """
    Алгоритм двойственного симплекс-метода.

    INPUT:
    - c
    - A
    - b
    - B

    OUTPUT: (dual symplex method bundle dict)
    - iter_num: number of iterations made
    - infeasible: True if the LPP is found to be infeasible
    - solved: True, если решение ЗЛП найдено
    - kappa: оптимальный план для прямой ЗЛП
    - y: оптимальный план для двойственной ЗЛП
    - B: итоговое множество базисных индексов
    """

    c = copy.deepcopy(c)
    A = copy.deepcopy(A)
    b = copy.deepcopy(b)
    B = copy.deepcopy(B)

    Ab_inv = None
    index = None
    for i in range(MAX_ITER):
        (solved, infeasible, kappa, y,
            Ab_inv, index) = iteration(c, A, b, B, Ab_inv, index)
        if infeasible:
            return {
                'iter_num': i+1,
                'infeasible': True,
                'solved': True,
            }
        if solved:
            return {
                'iter_num': i+1,
                'infeasible': False,
                'solved': True,
                'kappa': kappa,
                'y': y,
                'B': B,
            }

    # the "I give up" result
    return {
        'iter_num': i+1,
        'infeasible': False,
        'solved': False,
        'kappa': kappa,
        'y': y,
        'B': B,
    }



from distutils.core import setup

setup(
  name = 'rsksymplexmethod',
  packages = ['rsksymplexmethod'],
  version = '0.1',
  license='MIT',
  description = 'Simplex method by rasteeslove.',
  author = 'Rostislav Kayko',
  author_email = 'krastsislau@gmail.com',
  url = 'https://github.com/rasteeslove/rsksymplexmethod',
  download_url = 'https://github.com/rasteeslove/rsksymplexmethod/archive/v_01.tar.gz',
  keywords = ['Math', 'Simplex', 'Method'],
  install_requires=[
          'numpy',
          'scipy',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.10',
  ],
)
